# PWBus - RabbitmqEngine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 30 15:08:11 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine import _Engine
from pwbus.engines.rabbitmq.engine_thread import RabbitmqEngineThread


class RabbitmqEngine(_Engine):

    def __init__(self):
        super().__init__(
            name='rabbitmq'
        )

    # RabbitmqEngine.serve
    #
    def serve(self, channel_registry):
        return self.run(RabbitmqEngineThread, channel_registry)



# __init__.py
from .clients import *
from .commons import *
from .configs import *
from .connectors import *
from .engines import *
from .flows import *
from .tasks import *
from .transformations import *



# PWBus - MongoFlowIn & MongoFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec 14 11:32:03 -03 2019

from bson import Binary, Code, ObjectId
from bson.json_util import dumps
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.mongo import MongoConnectorIn, MongoConnectorOut


# MongoFlowIn
#
#
class MongoFlowIn(FlowIn):

    def __init__(self, request=None, host=None, port=None):
        super().__init__('in', MongoConnectorIn(host=host, port=port))
        self.host = host
        self.port = port

    # MongoFlowIn.receive
    #
    def receive(self, timeout=60):
        DEBUG = self.isDebugEnabled()
        resource_name = self.getRecvFrom().split('.')
        try:
            database = resource_name[0]
            collection = resource_name[1]
        except:
            log_debug(
                f'🟥 MongoFlowIn.receive - Invalid resource name [{self.getRecvFrom()}] specify "database.collection"')
            return None

        if DEBUG:
            log_debug(
                f'MongoFlowIn.receive - Started to [{resource_name}] format={self.getPayloadFormat()}')

        while True:
            try:
                if DEBUG:
                    log_debug(
                        "MongoFlowIn.receive.message - Getting connection")

                with self.getConnector().getConnectionPool().item() as connection:

                    db = connection[database]

                    while True:
                        if collection in db.list_collection_names():
                            break
                        log_debug(
                            f'MongoFlowIn.receive.message - Collection [{collection}] not exists in database [{database}] - Waiting ... (30s)')
                        sleep(30)

                    wait = 0

                    # pooling for wait messages
                    while True:

                        if wait <= 0:
                            wait = timeout
                            if DEBUG:
                                log_debug(
                                    f'⏱️  MongoFlowIn.receive.message - Waiting for messages from [{resource_name}|PSEUDOSYNC] ...')

                        response = db[collection].find_one_and_delete({}, {
                                                                      "_id": 0})

                        if response:
                            payload_in = dumps(response)
                            if DEBUG:
                                log_debug(
                                    f'↘️  MongoFlowIn.receive.message - Message received successfully from [{resource_name}] - payload_in [{payload_in}]')
                            return payload_in

                        sleep(0.100)
                        wait -= 0.100

            except KeyboardInterrupt:
                log_debug('MongoFlowIn.receive.message - Interrupted by user')
                self.stop()

            except:
                log_error(
                    traceback, f'MongoFlowIn.receive.message - Failed on receive message from [{self.getRecvFrom()}]')

            return None


# MongoFlowOut
#
#
class MongoFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', MongoConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # MongoFlowOut.send
    #
    def send(self):
        DEBUG = self.isDebugEnabled()

        resource_name = self.getSendTo().split('.')
        payload_out = self.getMessage().getPayload()
        try:
            database = resource_name[0]
            collection = resource_name[1]
        except:
            log_debug(
                f'🟥 MongoFlowOut.send - Invalid resource name [{self.getSendTo()}] specify "database.collection"')
            return False

        if DEBUG:
            log_debug(
                f'MongoFlowOut.send - Started to [{resource_name}] - payload_out [{payload_out}]')

        try:
            with self.getConnector().getConnectionPool().item() as connection:

                db = connection[database]
                col = db[collection]

                if col.insert(payload_out):
                    if DEBUG:
                        log_debug(
                            f'↗️  MongoFlowOut.send - Record inserted into mongodb')
                    return True
        except:
            log_error(
                traceback, f'MongoFlowOut.send - Failed to send the message to [{resource_name}]')
            traceback.print_exc()

        return False



# PWBus - Registry Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

from json import loads
import traceback

from pwbus.commons.logging import *
from pwbus.connectors.redis import RedisConnector

# Registry. - Get registry channel settings
#
#


class Registry(RedisConnector):

    # Registry.getChannel
    #
    def getChannel(self, channel):
        try:
            with self.getConnectionPool().item() as connection:
                return loads(connection.get(f'wpbus.registry.{channel}'))
        except:
            log_debug(f'Registry.getChannel - Channel [{channel}] not found')
            return None



### PWBus - TaskUniqueId
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Dec 15 10:36:07 -03 2019

from pwbus.commons.logging import *
from pwbus.connectors.redis import RedisConnector


class TaskUniqueId(RedisConnector):

    # TaskUniqueId.get
    #
    def get(self):
        try:
            with self.getConnectionPool().item() as connection:
                return connection.incr("pwbus_unique", amount=1)
        except:
            log_error("TaskUniqueId.get - Failed on get unique number")



# PWBus - MongoClient Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sun Nov 17 20:10:24 -03 2019

import boto3
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.clients.client import Client
from pwbus.connectors.mongo import MongoConnector

# MongoClient
#
#


class MongoClient(MongoConnector, Client):

    def __init__(self, host=None, port=None):
        super().__init__()

    # MongoClient.get
    #
    def get(self, resource_name, correlation_id, wait=0.05, retries=20):
        resource_name = resource_name.split('.')
        database = resource_name[0]
        collection = resource_name[1]
        log_debug(
            f'🔎 MongoClient.get - Retrieving response from [{resource_name}] with correlation_id [{correlation_id}] wait={wait}ms retries={retries}')
        try:
            with self.getConnectionPool().item() as connection:

                db = connection[database]
                retry_count = 0

                # pooling for wait messages
                while retry_count < retries:
                    response = db[collection].find_one_and_delete(
                        {"Pwbus-Correlation-Id": correlation_id},
                        {"_id": 0}
                    )

                    if response:
                        log_debug(
                            f'MongoClient.get - Response found - correlation_id [{correlation_id}]]')
                        return self.clear_header(response)
                    else:
                        sleep(wait)
                        retry_count += 1

                log_debug(
                    f'MongoClient.get - Response not found - correlation_id [{correlation_id}]')
                return None

        except:
            log_error(
                traceback, f'MongoClient.get - Failed to get message by correlation_id')

        return None



# PWBus - RabbitmqConnector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 30 15:01:58 -03 2019

import pika
from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.connectors._connector import _Connector

# RabbitmqConnectorIn
#
#


class RabbitmqConnector(_Connector):

    def __init__(self, host='rabbitmq', port=5672):
        super().__init__('rabbitmq', host, port)

        self.setConnectionPool(
            self.create_connection,
            max_size=10,
            max_usage=1000,
            idle=300,
            ttl=600
        )

    # RabbitmqConnector.create_connection
    #
    def create_connection(self):
        try:
            credentials = pika.PlainCredentials('guest', 'guest')
            return pika.BlockingConnection(pika.ConnectionParameters(
                host=self.host,
                port=self.port,
                virtual_host='/',
                credentials=credentials,
                blocked_connection_timeout=60
            ))
        except:
            log_error(
                traceback, 'RabbitmqConnector - Error on try connect to RABBITMQ server. Waiting 5s too retry...')
            sleep(5)


# RabbitmqConnectorIn
#
#
class RabbitmqConnectorIn(RabbitmqConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)

# RabbitmqConnectorOut
#
#


class RabbitmqConnectorOut(RabbitmqConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)



# PWBus - RedisConnector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import redis
import traceback

from pwbus.commons.logging import *
from pwbus.connectors._connector import _Connector

# RedisConnectorIn
#
#


class RedisConnector(_Connector):

    def __init__(self, host='redis', port=6379):
        super().__init__('redis', host=host, port=port)
        self.setConnectionPool(
            self.create_connection,
            max_size=100,
            max_usage=10000,
            idle=300,
            ttl=600
        )

    # RedisConnector.create_connection
    #
    def create_connection(self):
        try:
            return redis.Redis(
                host=self.host,
                port=self.port,
                db=0,
                password=None,
                socket_timeout=600
            )
        except:
            log_error(
                traceback, 'RedisConnector - Error on try connect to REDIS server')


# RedisConnectorIn
#
#
class RedisConnectorIn(RedisConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)

# RedisConnectorOut
#
#


class RedisConnectorOut(RedisConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)



# PWBus - MongoEngine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec 14 15:32:34 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine import _Engine
from pwbus.engines.mongo.engine_thread import MongoEngineThread


class MongoEngine(_Engine):

    def __init__(self):
        super().__init__(
            name='mongo'
        )

    # MongoEngine.serve
    #
    def serve(self, channel_registry):
        return self.run(MongoEngineThread, channel_registry)



# PWBus - SqsConnector Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec  7 09:46:09 -03 2019

import boto3
import traceback

from pwbus.commons.logging import *
from pwbus.connectors._connector import _Connector

# SqsConnectorIn
#
#


class SqsConnector(_Connector):

    def __init__(self, host=None, port=None):
        super().__init__('sqs', host, port)
        self.setConnectionPool(
            self.create_connection,
            max_size=10,
            max_usage=1000,
            idle=300,
            ttl=600
        )

    # SqsConnector.create_connection
    #
    def create_connection(self):
        try:
            return boto3.client('sqs')
        except:
            log_error(
                traceback, 'SqsConnector - Error on try connect to SQS server')


# SqsConnectorIn
#
#
class SqsConnectorIn(SqsConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)

# SqsConnectorOut
#
#


class SqsConnectorOut(SqsConnector):

    def __init__(self, host, port):
        super().__init__(host=host, port=port)



# PWBus - RedisEngine Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine import _Engine
from pwbus.engines.redis.engine_thread import RedisEngineThread


class RedisEngine(_Engine):

    def __init__(self):
        super().__init__(
            name='redis'
        )

    # RedisEngine.serve
    #
    def serve(self, channel_registry):
        return self.run(RedisEngineThread, channel_registry)



# PWBus - HttpFlowIn & HttpFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import requests
from requests.exceptions import HTTPError
import secrets
import traceback

from pwbus.commons.logging import *
from pwbus.flows._message import _Message
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.http import HttpConnectorIn, HttpConnectorOut

# HttpFlowIn
#
#


class HttpFlowIn(FlowIn):

    def __init__(self, request, host=None, port=None):
        super().__init__('in', HttpConnectorIn(host=host, port=port))
        self.request = request
        self.host = host
        self.port = port

    # HttpFlowIn.receive
    #
    def receive(self):

        if self.isDebugEnabled():
            log_debug(
                f'HttpFlowIn.receive - Started to [{self.getRecvFrom()}] format={self.getPayloadFormat()}')

        try:
            if self.request.content_type == 'application/json':
                payload_in = self.request.get_json()
            else:
                payload_in = self.request.get_data()

            for name, value in self.request.headers:
                if name in self.ALLOWED_HEADERS:
                    self.setHeadersEntry(name, value)

            if self.isDebugEnabled():
                log_debug(
                    f'↘️  HttpFlowIn.receive - Message received successfully from [{self.getRecvFrom()}]')
            return payload_in

        except:
            log_error(
                traceback, "HttpFlowIn.receive - Failed to receive message from '" + self.getRecvFrom() + "'")
            self.setResponse({"status_code": 500})
            return False

# HttpFlowOut
#
#


class HttpFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', HttpConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # HttpFlowOut.send
    #
    def send(self):
        if self.isDebugEnabled():
            log_debug(f'HttpFlowOut.send - Started to [{self.getSendTo()}]')

        try:
            if self.getPayloadFormat() == 'json':
                response = requests.post(
                    self.getSendTo(),
                    json=self.getMessage().getPayload(),
                    headers=self.getHeaders()
                )
            else:
                response = requests.post(
                    self.getSendTo(),
                    data=self.getMessage().getPayloadStream(),
                    headers=self.getHeaders()
                )

            if response.status_code != 200:
                if self.isDebugEnabled():
                    log_debug(
                        f'HttpFlowOut.send - status_code={response.status_code}')
                return False

            if 'Content-Type'in response.headers:
                if response.headers['Content-Type'] == 'application/json':
                    self.setResponse(response.json())
                else:
                    self.setResponse(response.text)

            if self.isDebugEnabled():
                log_debug("↗️  HttpFlowOut.send - Message sent successfully - task_id=" +
                          self.getMessage().getTaskId())
            return True

        except:
            log_debug(
                f'HttpFlowOut.send - Failed to send the message to [{self.getSendTo()}]')

        self.setHeadersEntry("Pwbus-Status-Code", 500)

        return False



# PWBus - Logging library
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 08:49:20 -03 2019

from datetime import datetime
from time import time
from threading import current_thread

from pwbus.engines.engine_monitor_event import EngineMonitorEvent

from sys import stdout, exit

LOG_DEBUG = True
LOG_ERROR = True

# logging.print_datetime
#


def print_datetime(arg1, arg2=''):
    print(current_thread().getName(), datetime.now().strftime(
        '%Y-%m-%d %H:%M:%S.%f'), f'{arg1} {arg2}')

# logging.getMillis
#


def getMillis():
    return int(round(time() * 1000))

# logging.log_debug
#


def log_debug_success(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('SUCES$$ 👍 ', arg1)
        if arg2 != '':
            print_datetime('     $$', arg2)
        print("\n")
        stdout.flush()

# logging.log_debug
#


def log_debug(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('Debug::', arg1)
        if arg2 != '':
            print_datetime('     ::', arg2)
        stdout.flush()

# logging.log_debug_var
#


def log_debug_var(key, value, arg1=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime(f'Debug>> {key} = [{value}]')
        if arg1 != '':
            print_datetime('     >>', arg1)
        stdout.flush()

# logging.log_error
#


def log_error(traceback, message):
    global LOG_ERROR
    EngineMonitorEvent().pushEvent(
        {"type": "error", "message": message, "exception": traceback.format_exc()})
    if LOG_DEBUG:
        formatted_lines = traceback.format_exc().splitlines()
        print_datetime(
            f'ERROR## 🟥 {message}\n     ## Message: {formatted_lines[-1]}')
        traceback.print_exc()
        stdout.flush()

# logging.log_fatal
#


def log_fatal(message):
    EngineMonitorEvent().pushEvent(
        {"type": "fatal", "message": message, "exception": ""})
    print_datetime('FATAL!! 🟥 Exiting with fatal error -', message)
    exit()


# logging.log_message_dump
#
def log_message_dump(message, message_dump):
    global LOG_DEBUG
    if LOG_DEBUG:
        if message_dump:
            print_datetime(f'DUMP .. {message} - message: {message_dump}')
        else:
            print_datetime(f'DUMP .. {message} - empty message')
        stdout.flush()

# logging.log_exit
#


def log_exit(arg1):
    print_datetime('EXIT !! 🟥 ', arg1)
    exit()

# logging.log_warn
#


def log_warn(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('WARN:: ⚠️ ', arg1)
        if arg2 != '':
            print_datetime('     ::', arg2)
        stdout.flush()

